type sched_t
type low_evt
type 'a evt_t=low_evt * 'a ref
type t

external cooperate: unit -> unit = "cooperate"
external low_create_fthread : sched_t -> (unit -> unit) -> t = "low_create_fthread"
external create_scheduler : unit -> sched_t = "create_scheduler"
external start_scheduler : sched_t -> unit = "start_scheduler"
external exit : unit -> unit = "exit_fthread"
external init : unit -> unit = "init"
external link_to : sched_t -> unit = "link_to"
external low_await : low_evt -> int -> unit = "low_fthread_await"
external low_signal : low_evt -> unit = "low_fthread_signal"
external low_create_event : unit -> low_evt = "fthread_create_event"
external alive : unit -> int = "alive"
external suspend_scheduler : sched_t -> unit = "suspend_scheduler"
external resume_scheduler : sched_t -> unit = "resume_scheduler"
external do_nothing : int -> unit = "do_nothing"
external register_main_scheduler : sched_t -> unit = "register_main_sched"
external decr_threads_alive : unit -> unit = "decr_threads_alive"
external incr_threads_alive : unit -> unit = "incr_threads_alive"

let rec remove_list a l=
  match !l with
    [] -> ()
  | x :: rl ->
      if x = a then
	l:=rl
      else remove_list a (ref rl)
;;

let main_scheduler=create_scheduler()

let unlink()=link_to main_scheduler

let create_event= 
  ((fun evt_val -> (low_create_event(),ref evt_val)) : 'a -> 'a evt_t)

let signal=
  ((fun (evt_id,evt_val) evt_new_val ->
    evt_val:=evt_new_val ; low_signal evt_id) : 'a evt_t -> 'a -> unit)

let awaitn n (evt_id,evt_val)=
  low_await evt_id n ; !evt_val

let link=link_to

let create sched f arg=
  ((low_create_fthread sched (fun () -> Gc.minor(); f arg)) : t)

let main_fthread=create main_scheduler (fun x -> ()) ()

let fd_list=ref []

let read_line=
  let rec loop alive=
    let (ready,_,_)=Unix.select [Unix.stdin] [] [] (0.0) in
    if ready=[] then 
      begin 
	if alive then
	  begin
	    fd_list:=Unix.stdin :: !fd_list;
	    decr_threads_alive();
	  end;
	cooperate();
	loop false
      end
    else
      begin
	incr_threads_alive();
	remove_list Unix.stdin fd_list;
	Pervasives.read_line()
      end
  in
  fun () -> loop true

let read fd buf off len=
  let rec loop alive=
    let (ready,_,_)=Unix.select [Unix.stdin] [] [] (0.0) in
    if ready=[] then 
      begin 
	if alive then
	  begin
	    fd_list:=fd :: !fd_list;
	    decr_threads_alive();
	  end;
	cooperate();
	loop false
      end
    else
      begin
	incr_threads_alive();
	remove_list fd fd_list;
	Unix.read fd buf off len
      end
  in
  fun () -> loop true


let _=
  register_main_scheduler main_scheduler;
  init();
  start_scheduler main_scheduler
