#include <unistd.h>
#include <sys/mman.h>
#include "migrate.h"

extern int threads_alive;
extern int fthread_mode;

sched_t sched_list=NULL;

struct sched_st dist_sched_st;
sched_t dist_sched=&dist_sched_st;
fthread_t fthread_tab[MAX_STATIC_FTHREAD];
int fthread_marks_tab[MAX_STATIC_FTHREAD];
int fthread_free_tab[MAX_STATIC_FTHREAD];
int free_index=0;


value get_scheduler(value unit)
{
  return (value)sched_list->current->sched;
}

void to_next_thread()
{
  sched_list->current=sched_list->current->next;

  while(sched_list->current->towait > 0){ // A optimiser
 
    sched_list->current->towait--;
    sched_list->current=sched_list->current->next;
  }
}

/** Calcule le prochain scheduler actif */

void to_next_scheduler()
{
  sched_t first=sched_list;;

 do{ 

    sched_list=sched_list->next;
  }while(((sched_list->status == 0) ||
	 (sched_list->current == NULL)) &&
	 (sched_list != first));

 if((sched_list == first) &&
    (sched_list->current == NULL)){
   exit(0);
 }
}

/** Le principe est simple : faire avancer le scheduler courant d'un Fair 
   Thread puis passer la main au scheduler suivant.*/

void low_cooperate()
{
  fthread_t curr=sched_list->current;

  if(!fthread_mode)
    failwith("Main thread can't cooperate !!");

  to_next_thread();
  to_next_scheduler();

  Before_switch(curr);
  swapcontext(curr->context,sched_list->current->context);
  After_switch(sched_list->current);
}


value reset_fthread(value unit)
{
  caml_bottom_of_stack = NULL;
  caml_exception_pointer = NULL;
  local_roots = NULL;
}



/** Fonction quivalente  la premire mais interfaable avec Caml. */

value cooperate(value unit)
{
  low_cooperate();
}

/** Le fait de commencer avec After_switch initialise le contexte GC 
   courant. */

void apply_callback()
{

  After_switch(sched_list->current);

  threads_alive--;

  callback((value)(sched_list->current->arg_pass), Val_unit);

  {
    fthread_t me=sched_list->current;

    if(me != me->next){
      me->prev->next=me->next;
      me->next->prev=me->prev;
      to_next_thread();
    }
    else{
      sched_list->current=NULL;
      sched_list->status=0;
    }

    to_next_scheduler(); 

    if(me->home != -1){
      fthread_free_tab[free_index]=me->home;
      free_index++;
    }

  //    free(me);   // rcupration de la mmoire
    if(me != sched_list->current)
      setcontext(sched_list->current->context); // on passe la main au suivant
    
  }
}

/** Fonction qui initialise un fair thread */

void init_fthread(fthread_t new_fthread)
{
  new_fthread->towait=0;
  new_fthread->bottom_of_stack = NULL;
  new_fthread->exception_pointer = NULL;
  new_fthread->local_roots = NULL;

}
  
void create_fthread(fthread_t new_fthread,sched_t sched,value f)
{ 
  static int id=1;

  if(sched->status == 1){
    threads_alive++;
  }

  sched->alive_ft++;
  init_fthread(new_fthread);

  new_fthread->id=id;
  new_fthread->sched=sched;
 
  if(sched->current == NULL){
    new_fthread->prev=new_fthread;
    new_fthread->next=new_fthread;
    sched->current=new_fthread;
  }
  else{
    new_fthread->next=sched->current;
    new_fthread->prev=sched->current->prev;
    sched->current->prev->next=new_fthread;
    sched->current->prev=new_fthread;
  }

  new_fthread->arg_pass=f;
  // Le code excut par le Fair Thread ne doit pas tre rcupr !!!!
  register_global_root(&(new_fthread->arg_pass)); 
  new_fthread->arg_pass=f;

#ifdef PPC
  setjmp(new_fthread->context);
  ((int*)new_fthread->context)[0]=
    (unsigned int)(new_fthread->stack)+STACK_SIZE;
  ((int*)new_fthread->context)[20]=(unsigned int)apply_callback+12;
  ((int*)new_fthread->context)[21]=(unsigned int)apply_callback+12;
#else

  setjmp(new_fthread->context);
  new_fthread->context->__jmpbuf[JB_SP]=
    (unsigned int)(new_fthread->stack)+STACK_SIZE;
  new_fthread->context->__jmpbuf[JB_PC]=(unsigned int)apply_callback;
#endif
  id++;
}

value low_create_fthread(value sched_p,value f)
{
  CAMLparam2(sched_p,f);

  fthread_t ft=malloc(sizeof(struct fthread_st));
  ft->home=-1;

  create_fthread(ft,(sched_t)sched_p,f);

  CAMLreturn((value)ft);
}


value create_static_fthread(value f,value base_p)
{
  CAMLparam2(f,base_p);
  static int curr=0;
  int home=(Int_val(base_p))*MAX_FTHREAD+curr;

  if(curr >= MAX_FTHREAD)
    failwith("Too many migrating threads\n");

  fthread_tab[home]=malloc(sizeof(struct fthread_st));

  if(fthread_tab[home] == (void*)-1)
    failwith("Couldn't address the fair thread\n");

  fthread_marks_tab[home]=1;

  create_fthread(fthread_tab[home],dist_sched,f);
  fthread_tab[home]->home=home;
  
  curr++;
  
  CAMLreturn((value)fthread_tab[home]);
}

value create_slot(value f,value home_p)
{
  CAMLparam2(f,home_p);

  int home=Int_val(home_p);

  if(!fthread_marks_tab[home])
    fthread_tab[home]=malloc(sizeof(struct fthread_st));
      /*mmap((void*)((home+1)*sizeof(struct fthread_st)),sizeof(struct fthread_st),
	   PROT_READ | PROT_WRITE, 
	   MAP_SHARED | MAP_ANONYMOUS | MAP_FIXED, 0 , 0);*/

  if(fthread_tab[home] == (void*)-1)
    failwith("Couldn't address the fair thread\n");

  create_fthread(fthread_tab[home],dist_sched,f);
  fthread_tab[home]->home=home;
    
  CAMLreturn((value)fthread_tab[home]);
}


int init_scheduler(sched_t sched)
{

  if(sched_list == NULL){
    sched->next=sched;
    sched->prev=sched;
    sched_list=sched;
  }
  else{
    sched->next=sched_list;
    sched->prev=sched_list->prev;
    sched_list->prev->next=sched;
    sched_list->prev=sched;
  }

  sched->alive_ft=0;
  sched->status=0;
  sched->current=NULL;
}

value create_scheduler(value unit)
{
  CAMLparam1(unit);
  sched_t sched=malloc(sizeof(struct sched_st));
  init_scheduler(sched);
  CAMLreturn((value)sched);
}

value start_scheduler(value sched)
{
  CAMLparam1(sched);

  if(((sched_t)sched)->status != 1){
  threads_alive+=((sched_t)sched)->alive_ft;
  ((sched_t)sched)->status=1;
  }

  CAMLreturn0;
}


value exit_fthread(value unit)
{
  fthread_mode=1;

  setcontext(sched_list->current->context);
}


value low_unlink(value unit)
{
  fthread_t me=sched_list->current;
  sched_list->alive_ft--;

  if(me != me->next){
    me->prev->next=me->next;
    me->next->prev=me->prev;
    to_next_thread();
  }
    else{
      sched_list->current=NULL;
      sched_list->status=0;
    }

  return Val_unit;
}

value low_link(value sched_p,value fth)
{
  fthread_t curr=(fthread_t)fth;
  sched_t sched=(sched_t)sched_p;

  sched->alive_ft++;

  if(sched->current == NULL){
    curr->prev=curr;
    curr->next=curr;
    sched->current=curr;
  }
  else{
    curr->next=sched->current;
    curr->prev=sched->current->prev;
    sched->current->prev->next=curr;
    sched->current->prev=curr;
  }
  return Val_unit;
}


value link_to(value sched)
{
  CAMLparam1(sched);
  fthread_t curr=sched_list->current;
  
  if(!fthread_mode)
    failwith("Can't unlink the main thread !!");
  
  low_unlink(0);
  low_link(sched,(value)curr);

  to_next_scheduler();

  Before_switch(curr);
  swapcontext(curr->context,sched_list->current->context);
  After_switch(sched_list->current);
   
  CAMLreturn0;
}

/** Fonction de signalisation la file des Fthread est scanne 
    et chaque Fair Thread en attente de l'vnement pass en 
    paramtre est rveill avec la valeur event_val mise  jour.*/

value low_fthread_signal(value event_id)
{
  CAMLparam1(event_id);

  fthread_t ptr=sched_list->current;

  if(!fthread_mode)
    failwith("The main thread can't signal !!");

  do{
    if(ptr->event_id == event_id){
      ptr->towait=0;
    }
    ptr=ptr->next;
  }while(ptr != sched_list->current);

  CAMLreturn0;
}


value low_fthread_await(value event_id,value towait)
{
  CAMLparam2(event_id,towait);
  CAMLlocal1(res);

  if(!fthread_mode)
    failwith("The main thread can't await\n");

  res=(value)0;

  if(Int_val(towait) > 0){
    sched_list->current->towait=Int_val(towait);
    sched_list->current->event_id=event_id;    
    low_cooperate();
  }
  else sched_list->current->towait=0;

  CAMLreturn0;
}

value fthread_create_event(value unit)
{
  CAMLparam1(unit);
  static int count=0;
  
  count++;

  CAMLreturn(Val_int(count));
}

value get_threads_alive(value unit)
{
  CAMLparam1(unit);
  CAMLreturn(Val_int(threads_alive));
}

value suspend_scheduler(value sched)
{
  CAMLparam1(sched);
  if(((sched_t)sched)->status != 0){
    threads_alive-=((sched_t)sched)->alive_ft;
    ((sched_t)sched)->status=0;
  }
  CAMLreturn0;
}

value resume_scheduler(value sched)
{
  CAMLparam1(sched);
  if(((sched_t)sched)->status != 1){
    threads_alive+=((sched_t)sched)->alive_ft;
    ((sched_t)sched)->status=1;
  }
  CAMLreturn0;
}



